// Copyright 2020 TiKV Project Authors. Licensed under Apache-2.0.

use std::{
    cmp,
    cmp::Reverse,
    collections::{BTreeMap, BinaryHeap},
    sync::Arc,
};

use collections::{HashMap, HashSet};
use raftstore::store::RegionReadProgress;
use txn_types::TimeStamp;

use crate::metrics::RTS_RESOLVED_FAIL_ADVANCE_VEC;

#[derive(Debug, Clone, Copy)]
pub struct ResolvedTs {
    pub raw_ts: TimeStamp,
    pub txn_ts: TimeStamp,
}

impl ResolvedTs {
    pub fn default() -> ResolvedTs {
        ResolvedTs {
            raw_ts: TimeStamp::zero(),
            txn_ts: TimeStamp::zero(),
        }
    }
    pub fn min(&self) -> TimeStamp {
        cmp::min(self.raw_ts, self.txn_ts)
    }

    pub fn is_min_ts_from_raw(&self) -> bool {
        self.raw_ts < self.txn_ts
    }
}

// Resolver resolves timestamps that guarantee no more commit will happen before
// the timestamp.
pub struct Resolver {
    region_id: u64,
    // key -> start_ts
    locks_by_key: HashMap<Arc<[u8]>, TimeStamp>,
    // start_ts -> locked keys.
    lock_ts_heap: BTreeMap<TimeStamp, HashSet<Arc<[u8]>>>,
    // raw ts, depend on "non-decreasing" of entries' timestamp in the same region.
    // BinaryHeap is max heap, so reverse order to get a min heap. Only used in rawkv.
    raw_lock_ts_heap: BinaryHeap<Reverse<TimeStamp>>,
    // The timestamps that guarantees no more commit will happen before.
    resolved_ts: ResolvedTs,
    // The highest index `Resolver` had been tracked
    tracked_index: u64,
    // The region read progress used to utilize `resolved_ts` to serve stale read request
    read_progress: Option<Arc<RegionReadProgress>>,
    // The timestamps that advance the resolved_ts when there is no more write.
    min_ts: TimeStamp,
    // Whether the `Resolver` is stopped
    stopped: bool,
}

impl std::fmt::Debug for Resolver {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let far_lock = self.lock_ts_heap.iter().next();
        let mut dt = f.debug_tuple("Resolver");
        dt.field(&format_args!("region={}", self.region_id));

        if let Some((ts, keys)) = far_lock {
            dt.field(&format_args!(
                "far_lock={:?}",
                keys.iter()
                    // We must use Display format here or the redact won't take effect.
                    .map(|k| format!("{}", log_wrappers::Value::key(k)))
                    .collect::<Vec<_>>()
            ));
            dt.field(&format_args!("far_lock_ts={:?}", ts));
        }

        dt.finish()
    }
}

impl Resolver {
    pub fn new(region_id: u64) -> Resolver {
        Resolver::with_read_progress(region_id, None)
    }

    pub fn with_read_progress(
        region_id: u64,
        read_progress: Option<Arc<RegionReadProgress>>,
    ) -> Resolver {
        Resolver {
            region_id,
            resolved_ts: ResolvedTs::default(),
            locks_by_key: HashMap::default(),
            lock_ts_heap: BTreeMap::new(),
            raw_lock_ts_heap: BinaryHeap::new(),
            read_progress,
            tracked_index: 0,
            min_ts: TimeStamp::zero(),
            stopped: false,
        }
    }

    pub fn resolved_ts(&self) -> TimeStamp {
        self.resolved_ts.min()
    }

    pub fn size(&self) -> usize {
        self.locks_by_key.keys().map(|k| k.len()).sum::<usize>()
            + self
                .lock_ts_heap
                .values()
                .map(|h| h.iter().map(|k| k.len()).sum::<usize>())
                .sum::<usize>()
    }

    pub fn locks(&self) -> &BTreeMap<TimeStamp, HashSet<Arc<[u8]>>> {
        &self.lock_ts_heap
    }

    pub fn stop_tracking(&mut self) {
        // TODO: should we also clear the lock heap?
        self.stopped = true;
        self.read_progress.take();
    }

    pub fn update_tracked_index(&mut self, index: u64) {
        assert!(
            self.tracked_index <= index,
            "region {}, tracked_index: {}, incoming index: {}",
            self.region_id,
            self.tracked_index,
            index
        );
        self.tracked_index = index;
    }

    pub fn track_lock(&mut self, start_ts: TimeStamp, key: Vec<u8>, index: Option<u64>) {
        if let Some(index) = index {
            self.update_tracked_index(index);
        }
        debug!(
            "track lock {}@{}, region {}",
            &log_wrappers::Value::key(&key),
            start_ts,
            self.region_id
        );
        let key: Arc<[u8]> = key.into_boxed_slice().into();
        self.locks_by_key.insert(key.clone(), start_ts);
        self.lock_ts_heap.entry(start_ts).or_default().insert(key);
    }

    pub fn untrack_lock(&mut self, key: &[u8], index: Option<u64>) {
        if let Some(index) = index {
            self.update_tracked_index(index);
        }
        let start_ts = if let Some(start_ts) = self.locks_by_key.remove(key) {
            start_ts
        } else {
            debug!("untrack a lock that was not tracked before"; "key" => &log_wrappers::Value::key(key));
            return;
        };
        debug!(
            "untrack lock {}@{}, region {}",
            &log_wrappers::Value::key(key),
            start_ts,
            self.region_id,
        );

        let entry = self.lock_ts_heap.get_mut(&start_ts);
        if let Some(locked_keys) = entry {
            locked_keys.remove(key);
            if locked_keys.is_empty() {
                self.lock_ts_heap.remove(&start_ts);
            }
        }
    }

    pub fn raw_track_lock(&mut self, ts: TimeStamp) {
        debug!("raw track ts {}, region {}", ts, self.region_id);
        self.raw_lock_ts_heap.push(Reverse(ts));
    }

    // untrack all timestamps smaller than input ts, depend on the raw ts in one
    // region is non-decreasing
    pub fn raw_untrack_lock(&mut self, ts: TimeStamp) {
        debug!("raw untrack ts before {}, region {}", ts, self.region_id);
        while let Some(&Reverse(min_ts)) = self.raw_lock_ts_heap.peek() {
            if min_ts > ts {
                break;
            }
            self.raw_lock_ts_heap.pop();
        }
    }

    /// Try to advance resolved ts.
    ///
    /// `min_ts` advances the resolver even if there is no write.
    /// Return None means the resolver is not initialized.
    pub fn resolve(&mut self, min_ts: TimeStamp) -> ResolvedTs {
        // The `Resolver` is stopped, not need to advance, just return the current
        // `resolved_ts`
        if self.stopped {
            return self.resolved_ts;
        }
        // Find the min start ts.
        let min_lock = self.lock_ts_heap.keys().next().cloned();
        let has_lock = min_lock.is_some();
        let min_start_ts = min_lock.unwrap_or(min_ts);

        // No more commit happens before the ts.
        let new_txn_resolved_ts = cmp::min(min_start_ts, min_ts);
        if self.resolved_ts.txn_ts >= new_txn_resolved_ts {
            let label = if has_lock { "has_lock" } else { "stale_ts" };
            RTS_RESOLVED_FAIL_ADVANCE_VEC
                .with_label_values(&[label])
                .inc();
        }

        // Resolved ts never decrease.
        self.resolved_ts.txn_ts = cmp::max(self.resolved_ts.txn_ts, new_txn_resolved_ts);

        // Publish an `(apply index, safe ts)` item into the region read progress
        if let Some(rrp) = &self.read_progress {
            rrp.update_safe_ts(self.tracked_index, self.resolved_ts.txn_ts.into_inner());
        }

        let min_raw_ts = self
            .raw_lock_ts_heap
            .peek()
            .map_or(min_ts, |ts| ts.to_owned().0);
        // Resolved ts never decrease.
        self.resolved_ts.raw_ts = cmp::max(self.resolved_ts.raw_ts, min_raw_ts);

        let new_min_ts = if has_lock {
            // If there are some lock, the min_ts must be smaller than
            // the min start ts, so it guarantees to be smaller than
            // any late arriving commit ts.
            new_txn_resolved_ts // cmp::min(min_start_ts, min_ts)
        } else {
            min_ts
        };
        // Min ts never decrease.
        self.min_ts = cmp::max(self.min_ts, new_min_ts);

        self.resolved_ts
    }
}

#[cfg(test)]
mod tests {
    use txn_types::Key;

    use super::*;

    #[derive(Clone)]
    enum Event {
        // start_ts, key
        Lock(u64, Key),
        // key
        Unlock(Key),
        // raw ts
        RawLock(u64),
        // raw ts
        RawUnlock(u64),
        // min_ts, expect
        Resolve(u64, u64),
    }

    #[test]
    fn test_resolve() {
        let cases = vec![
            vec![Event::Lock(1, Key::from_raw(b"a")), Event::Resolve(2, 1)],
            vec![
                Event::Lock(1, Key::from_raw(b"a")),
                Event::Unlock(Key::from_raw(b"a")),
                Event::Resolve(2, 2),
            ],
            vec![
                Event::Lock(3, Key::from_raw(b"a")),
                Event::Unlock(Key::from_raw(b"a")),
                Event::Resolve(2, 2),
            ],
            vec![
                Event::Lock(1, Key::from_raw(b"a")),
                Event::Unlock(Key::from_raw(b"a")),
                Event::Lock(1, Key::from_raw(b"b")),
                Event::Resolve(2, 1),
            ],
            vec![
                Event::Lock(2, Key::from_raw(b"a")),
                Event::Unlock(Key::from_raw(b"a")),
                Event::Resolve(2, 2),
                // Pessimistic txn may write a smaller start_ts.
                Event::Lock(1, Key::from_raw(b"a")),
                Event::Resolve(2, 2),
                Event::Unlock(Key::from_raw(b"a")),
                Event::Resolve(3, 3),
            ],
            vec![
                Event::Unlock(Key::from_raw(b"a")),
                Event::Lock(2, Key::from_raw(b"a")),
                Event::Unlock(Key::from_raw(b"a")),
                Event::Unlock(Key::from_raw(b"a")),
                Event::Resolve(3, 3),
            ],
            vec![
                Event::Lock(2, Key::from_raw(b"a")),
                Event::Resolve(4, 2),
                Event::Unlock(Key::from_raw(b"a")),
                Event::Resolve(5, 5),
            ],
            // Rollback may contain a key that is not locked.
            vec![
                Event::Lock(1, Key::from_raw(b"a")),
                Event::Unlock(Key::from_raw(b"b")),
                Event::Unlock(Key::from_raw(b"a")),
            ],
            // raw track lock
            vec![Event::RawLock(1), Event::Resolve(2, 1)],
            vec![Event::RawLock(1), Event::RawUnlock(1), Event::Resolve(2, 2)],
            vec![Event::RawLock(1), Event::RawUnlock(2), Event::Resolve(5, 5)],
            vec![
                Event::RawLock(1),
                Event::RawUnlock(2),
                Event::RawLock(3),
                Event::Resolve(5, 3),
            ],
            vec![
                Event::RawLock(1),
                Event::RawUnlock(2),
                Event::RawLock(3),
                Event::RawLock(4),
                Event::Resolve(5, 3),
            ],
            // raw and txn mixed
            vec![
                Event::Lock(1, Key::from_raw(b"a")),
                Event::RawLock(2),
                Event::RawUnlock(3),
                Event::Resolve(5, 1),
                Event::Unlock(Key::from_raw(b"a")),
                Event::Resolve(6, 6),
            ],
            vec![
                Event::Lock(1, Key::from_raw(b"a")),
                Event::RawLock(2),
                Event::RawLock(3),
                Event::Resolve(5, 1),
                Event::Unlock(Key::from_raw(b"a")),
                Event::Resolve(6, 2),
            ],
        ];

        for (i, case) in cases.into_iter().enumerate() {
            let mut resolver = Resolver::new(1);
            for e in case.clone() {
                match e {
                    Event::Lock(start_ts, key) => {
                        resolver.track_lock(start_ts.into(), key.into_raw().unwrap(), None)
                    }
                    Event::Unlock(key) => resolver.untrack_lock(&key.into_raw().unwrap(), None),
                    Event::RawLock(ts) => resolver.raw_track_lock(ts.into()),
                    Event::RawUnlock(ts) => resolver.raw_untrack_lock(ts.into()),
                    Event::Resolve(min_ts, expect) => {
                        assert_eq!(
                            resolver.resolve(min_ts.into()).min(),
                            expect.into(),
                            "case {}",
                            i
                        )
                    }
                }
            }
        }
    }
}
